from snowddl.blueprint import SchemaBlueprint, DatabaseBlueprint
from snowddl.resolver.abc_resolver import AbstractResolver, ResolveResult, ObjectType, SnowDDLUnsupportedError


class SchemaResolver(AbstractResolver):
    def get_object_type(self) -> ObjectType:
        return ObjectType.SCHEMA

    def get_existing_objects(self):
        return self.engine.schema_cache.schemas

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(SchemaBlueprint)

    def create_object(self, bp: SchemaBlueprint):
        query = self.engine.query_builder()
        query.append("CREATE")

        if bp.is_transient:
            query.append("TRANSIENT")

        query.append("SCHEMA {full_name:i}", {
            "full_name": bp.full_name
        })

        query.append_nl("WITH MANAGED ACCESS")

        if bp.retention_time is not None:
            query.append_nl("DATA_RETENTION_TIME_IN_DAYS = {retention_time:d}", {
                "retention_time": bp.retention_time
            })

        if bp.comment:
            query.append_nl("COMMENT = {comment}", {
                "comment": bp.comment,
            })

        self.engine.execute_safe_ddl(query)

        return ResolveResult.CREATE

    def compare_object(self, bp: SchemaBlueprint, row: dict):
        if bp.is_transient != row['is_transient']:
            if bp.is_transient:
                raise SnowDDLUnsupportedError("Cannot change PERMANENT object into TRANSIENT object")
            else:
                raise SnowDDLUnsupportedError("Cannot change TRANSIENT object into PERMANENT object")

        query = self.engine.query_builder()

        query.append("ALTER SCHEMA {full_name:i} SET ", {
            "full_name": bp.full_name,
        })

        if bp.retention_time is not None and bp.retention_time != row['retention_time']:
            query.append_nl("DATA_RETENTION_TIME_IN_DAYS = {retention_time:d}", {
                "retention_time": bp.retention_time
            })

        if bp.comment != row['comment']:
            query.append_nl("COMMENT = {comment}", {
                "comment": bp.comment
            })

        if query.fragment_count() > 1:
            self.engine.execute_unsafe_ddl(query)
            return ResolveResult.ALTER

        return ResolveResult.NOCHANGE

    def drop_object(self, row: dict):
        self.engine.execute_unsafe_ddl("DROP SCHEMA {database:i}.{schema:i}", {
            "database": row['database'],
            "schema": row['schema'],
        })

        return ResolveResult.DROP

    def destroy(self):
        # Do nothing, all schemas are dropped automatically on DROP DATABASE
        pass

    def _post_process(self):
        for result in self.resolved_objects.values():
            if result != ResolveResult.NOCHANGE:
                # Reload cache if at least one object was changed
                self.engine.schema_cache.reload()
                break

    def _resolve_drop(self):
        # Drop existing schemas without blueprints
        # with additional check for "sandbox" database
        tasks = {}

        for full_name in sorted(self.existing_objects):
            if full_name not in self.blueprints and not self._is_sandbox_database(full_name):
                tasks[full_name] = (self.drop_object, self.existing_objects[full_name])

        self._process_tasks(tasks)

    def _is_sandbox_database(self, schema_full_name):
        database_full_name = '.'.join(schema_full_name.split('.')[:1])
        database_bp = self.config.get_blueprints_by_type(DatabaseBlueprint).get(database_full_name)

        if database_bp and database_bp.is_sandbox:
            return True

        return False



from snowddl.blueprint import SequenceBlueprint
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult, ObjectType


class SequenceResolver(AbstractSchemaObjectResolver):
    def get_object_type(self) -> ObjectType:
        return ObjectType.SEQUENCE

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta("SHOW SEQUENCES IN SCHEMA {database:i}.{schema:i}", {
            "database": schema['database'],
            "schema": schema['schema'],
        })

        for r in cur:
            existing_objects[f"{r['database_name']}.{r['schema_name']}.{r['name']}"] = {
                "database": r['database_name'],
                "schema": r['schema_name'],
                "name": r['name'],
                "owner": r['owner'],
                "next_value": r['next_value'],
                "interval": r['interval'],
                "comment": r['comment'] if r['comment'] else None,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(SequenceBlueprint)

    def create_object(self, bp: SequenceBlueprint):
        query = self.engine.query_builder()

        query.append("CREATE SEQUENCE {full_name:i}", {
            "full_name": bp.full_name,
        })

        query.append_nl("START = {start:d}", {
            "start": bp.start,
        })

        query.append_nl("INCREMENT = {interval:d}", {
            "interval": bp.interval,
        })

        if bp.comment:
            query.append_nl("COMMENT = {comment}", {
                "comment": bp.comment,
            })

        self.engine.execute_safe_ddl(query)

        return ResolveResult.CREATE

    def compare_object(self, bp: SequenceBlueprint, row: dict):
        result = ResolveResult.NOCHANGE

        if bp.interval != row['interval']:
            self.engine.execute_safe_ddl("ALTER SEQUENCE {full_name:i} SET INCREMENT = {interval:d}", {
                "full_name": bp.full_name,
                "interval": bp.interval,
            })

            result = ResolveResult.ALTER

        if bp.comment != row['comment']:
            self.engine.execute_safe_ddl("ALTER SEQUENCE {full_name:i} SET COMMENT = {comment}", {
                "full_name": bp.full_name,
                "comment": bp.comment,
            })

            result = ResolveResult.ALTER

        return result

    def drop_object(self, row: dict):
        self.engine.execute_unsafe_ddl("DROP SEQUENCE {database:i}.{schema:i}.{name:i}", {
            "database": row['database'],
            "schema": row['schema'],
            "name": row['name'],
        })

        return ResolveResult.DROP



from jsonschema import validate
from pathlib import Path
from re import compile, IGNORECASE
from typing import TYPE_CHECKING
from yaml import safe_load

if TYPE_CHECKING:
    from snowddl.parser.abc_parser import AbstractParser

class ParsedFile:
    placeholder_start = '${{ '
    placeholder_end = ' }}'
    placeholder_re = compile(r'\${{\s([a-z0-9._-]+)\s}}', IGNORECASE)

    def __init__(self, parser: "AbstractParser", path: Path, json_schema: dict):
        self.parser = parser

        self.path = path
        self.name = path.stem
        self.json_schema = json_schema

        self.database = None
        self.schema = None

        self._guess_database_schema_from_path()
        self._load_params()

        self._apply_placeholders(self.params)
        self._validate_json_schema()

    def _guess_database_schema_from_path(self):
        try:
            relative_path = self.path.relative_to(self.parser.base_path)
        except ValueError:
            return

        if len(relative_path.parts) > 1:
            self.database = relative_path.parts[0]

        if len(relative_path.parts) > 2:
            self.schema = relative_path.parts[1]

    def _load_params(self):
        with self.path.open('r', encoding='utf-8') as f:
            self.params = safe_load(f) or {}

    def _apply_placeholders(self, data: dict):
        for k, v in data.items():
            if isinstance(v, dict):
                self._apply_placeholders(v)
            elif isinstance(v, list):
                data[k] = [self._apply_placeholders_inner(i) if isinstance(i, str) else i for i in v]
            elif isinstance(v, str):
                data[k] = self._apply_placeholders_inner(v)

    def _apply_placeholders_inner(self, val: str):
        if self.placeholder_start in val:
            m = self.placeholder_re.fullmatch(val)

            if m:
                # Value is a single placeholder, return and preserve type
                return self.parser.config.get_placeholder(m.group(1).upper())
            else:
                # Value is a string with multiple placeholders or other parts, replace and return as string
                return self.placeholder_re.sub(lambda m: str(self.parser.config.get_placeholder(m.group(1).upper())), val)

        return val

    def _validate_json_schema(self):
        validate(self.params, self.json_schema)



from snowddl.blueprint import RoleBlueprint, UserBlueprint, Grant, build_role_ident
from snowddl.resolver.abc_role_resolver import AbstractRoleResolver, ObjectType


class UserRoleResolver(AbstractRoleResolver):
    def get_role_suffix(self):
        return self.config.USER_ROLE_SUFFIX

    def get_blueprints(self):
        blueprints = []

        for user in self.config.get_blueprints_by_type(UserBlueprint).values():
            blueprints.append(self.get_blueprint_user_role(user))

        return {str(bp.full_name): bp for bp in blueprints}

    def get_blueprint_user_role(self, user: UserBlueprint):
        grants = []

        for business_role in user.business_roles:
            grants.append(Grant(
                privilege="USAGE",
                on=ObjectType.ROLE,
                name=business_role,
            ))

        bp = RoleBlueprint(
            full_name=build_role_ident(self.config.env_prefix, user.full_name, self.get_role_suffix()),
            grants=grants,
            future_grants=[],
            comment=None,
        )

        return bp



from ._yaml import SnowDDLDumper, YamlFoldedStr, YamlLiteralStr
from .abc_converter import AbstractConverter, ConvertResult
from .abc_schema_object_converter import AbstractSchemaObjectConverter
from .database import DatabaseConverter
from .schema import SchemaConverter
from .sequence import SequenceConverter
from .table import TableConverter
from .view import ViewConverter


default_converter_sequence = [
    DatabaseConverter,
    SchemaConverter,
    SequenceConverter,
    TableConverter,
    #ViewConverter,
]



from snowddl.blueprint import StreamBlueprint, SchemaObjectIdent, ObjectType, build_schema_object_ident
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


stream_json_schema = {
    "type": "object",
    "properties": {
        "object_type": {
            "type": "string"
        },
        "object_name": {
            "type": "string"
        },
        "append_only": {
            "type": "boolean"
        },
        "insert_only": {
            "type": "boolean"
        },
        "show_initial_rows": {
            "type": "boolean"
        },
        "comment": {
            "type": "string"
        }
    },
    "required": ["object_type", "object_name"],
    "additionalProperties": False
}


class StreamParser(AbstractParser):
    def load_blueprints(self):
        self.parse_schema_object_files("stream", stream_json_schema, self.process_stream)

    def process_stream(self, f: ParsedFile):
        bp = StreamBlueprint(
            full_name=SchemaObjectIdent(self.env_prefix, f.database, f.schema, f.name),
            object_type=ObjectType[f.params['object_type']],
            object_name=build_schema_object_ident(self.env_prefix, f.params['object_name'], f.database, f.schema),
            append_only=f.params.get('append_only'),
            insert_only=f.params.get('insert_only'),
            show_initial_rows=f.params.get('show_initial_rows'),
            comment=f.params.get('comment'),
        )

        self.config.add_blueprint(bp)



from snowddl.blueprint import PipeBlueprint
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult, ObjectType
from snowddl.resolver._utils import coalesce


class PipeResolver(AbstractSchemaObjectResolver):
    def get_object_type(self) -> ObjectType:
        return ObjectType.PIPE

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta("SHOW PIPES IN SCHEMA {database:i}.{schema:i}", {
            "database": schema['database'],
            "schema": schema['schema'],
        })

        for r in cur:
            existing_objects[f"{r['database_name']}.{r['schema_name']}.{r['name']}"] = {
                "database": r['database_name'],
                "schema": r['schema_name'],
                "name": r['name'],
                "definition": r['definition'],
                "integration": r['integration'],
                "pattern": r['pattern'],
                "comment": r['comment'],
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(PipeBlueprint)

    def create_object(self, bp: PipeBlueprint):
        common_query = self._build_common_pipe_sql(bp)
        create_query = self.engine.query_builder()

        create_query.append("CREATE PIPE {full_name:i}", {
            "full_name": bp.full_name,
        })

        create_query.append(common_query)

        self.engine.execute_safe_ddl(create_query)

        self.engine.execute_safe_ddl("COMMENT ON PIPE {full_name:i} IS {comment}", {
            "full_name": bp.full_name,
            "comment": common_query.add_short_hash(bp.comment),
        })

        return ResolveResult.CREATE

    def compare_object(self, bp: PipeBlueprint, row: dict):
        common_query = self._build_common_pipe_sql(bp)

        if not common_query.compare_short_hash(row['comment']):
            replace_query = self.engine.query_builder()

            replace_query.append("CREATE OR REPLACE PIPE {full_name:i}", {
                "full_name": bp.full_name,
            })

            replace_query.append(common_query)

            self.engine.execute_unsafe_ddl(replace_query)

            self.engine.execute_unsafe_ddl("COMMENT ON PIPE {full_name:i} IS {comment}", {
                "full_name": bp.full_name,
                "comment": common_query.add_short_hash(bp.comment),
            })

            return ResolveResult.REPLACE

        return ResolveResult.NOCHANGE

    def drop_object(self, row: dict):
        self.engine.execute_unsafe_ddl("DROP PIPE {database:i}.{schema:i}.{name:i}", {
            "database": row['database'],
            "schema": row['schema'],
            "name": row['name'],
        })

        return ResolveResult.DROP

    def _build_common_pipe_sql(self, bp: PipeBlueprint):
        query = self.engine.query_builder()

        query.append_nl("AUTO_INGEST = {auto_ingest:b}", {
            "auto_ingest": bp.auto_ingest,
        })

        if bp.aws_sns_topic:
            query.append_nl("AWS_SNS_TOPIC = {aws_sns_topic}", {
                "aws_sns_topic": bp.aws_sns_topic,
            })

        if bp.integration:
            query.append_nl("INTEGRATION = {integration:i}", {
                "integration": bp.integration,
            })

        if bp.comment:
            query.append_nl("COMMENT = {comment}", {
                "comment": bp.comment,
            })

        query.append_nl("AS")
        query.append_nl(self._build_copy_into(bp))

        return query

    def _build_copy_into(self, bp: PipeBlueprint):
        query = self.engine.query_builder()

        query.append("COPY INTO {table_name:i}", {
            "table_name": bp.copy_table_name,
        })

        if bp.copy_transform:
            query.append(" (")

            for idx, col_name in enumerate(bp.copy_transform):
                query.append_nl("    {comma:r}{col_name:i}", {
                    "comma": "  " if idx == 0 else ", ",
                    "col_name": col_name,
                })

            query.append_nl(") FROM (")
            query.append_nl("    SELECT")

            for idx, col_expr in enumerate(bp.copy_transform.values()):
                query.append_nl("        {comma:r}{col_expr:r}", {
                    "comma": "  " if idx == 0 else ", ",
                    "col_expr": col_expr,
                })

            query.append_nl("    FROM @{stage_name:i}{path:r}", {
                "stage_name": bp.copy_stage_name,
                "path": coalesce(bp.copy_path, ''),
            })

            query.append_nl(")")
        else:
            query.append_nl("FROM @{stage_name:i}{path:r}", {
                "stage_name": bp.copy_stage_name,
                "path": coalesce(bp.copy_path, ''),
            })

        if bp.copy_pattern:
            query.append_nl("PATTERN = {pattern}", {
                "pattern": bp.copy_pattern,
            })

        if bp.copy_file_format:
            query.append_nl("FILE_FORMAT = {file_format:i}", {
                "file_format": bp.copy_file_format,
            })

        if bp.copy_options:
            for k, v in bp.copy_options.items():
                query.append_nl("{option_name:r} = {option_value:dp}", {
                    "option_name": k,
                    "option_value": v,
                })

        return query



from snowddl.blueprint import MaskingPolicyBlueprint, ObjectType, Edition, SchemaObjectIdent
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult


class MaskingPolicyResolver(AbstractSchemaObjectResolver):
    skip_on_empty_blueprints = True
    skip_min_edition = Edition.ENTERPRISE

    def get_object_type(self) -> ObjectType:
        return ObjectType.MASKING_POLICY

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta("SHOW MASKING POLICIES IN SCHEMA {database:i}.{schema:i}", {
            "database": schema['database'],
            "schema": schema['schema'],
        })

        for r in cur:
            full_name = f"{r['database_name']}.{r['schema_name']}.{r['name']}"

            existing_objects[full_name] = {
                "database": r['database_name'],
                "schema": r['schema_name'],
                "name": r['name'],
                "comment": r['comment'] if r['comment'] else None,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(MaskingPolicyBlueprint)

    def create_object(self, bp: MaskingPolicyBlueprint):
        self._create_policy(bp)
        self._apply_policy_refs(bp, skip_existing=True)

        return ResolveResult.CREATE

    def compare_object(self, bp: MaskingPolicyBlueprint, row: dict):
        cur = self.engine.execute_meta("DESC MASKING POLICY {full_name:i}", {
            "full_name": bp.full_name,
        })

        r = cur.fetchone()

        # If signature or return type was changed, policy and all references must be dropped and created again
        if r['signature'] != f"({', '.join([f'{a.name} {a.type.base_type.name}' for a in bp.arguments])})" \
        or r['return_type'] != str(bp.returns):
            self._drop_policy_refs(bp.full_name)
            self._drop_policy(bp.full_name)

            self._create_policy(bp)
            self._apply_policy_refs(bp, skip_existing=True)

            return ResolveResult.REPLACE

        result = ResolveResult.NOCHANGE

        if self._apply_policy_refs(bp):
            result = ResolveResult.ALTER

        if r['body'] != bp.body:
            self.engine.execute_unsafe_ddl("ALTER MASKING POLICY {full_name:i} SET BODY -> {body:r}", {
                "full_name": bp.full_name,
                "body": bp.body,
            }, condition=self.engine.settings.execute_masking_policy)

            result = ResolveResult.ALTER

        if row['comment'] != bp.comment:
            self.engine.execute_unsafe_ddl("ALTER MASKING POLICY {full_name:i} SET COMMENT = {comment}", {
                "full_name": bp.full_name,
                "comment": bp.comment,
            }, condition=self.engine.settings.execute_masking_policy)

            result = ResolveResult.ALTER

        return result

    def drop_object(self, row: dict):
        self._drop_policy_refs(SchemaObjectIdent('', row['database'], row['schema'], row['name']))
        self._drop_policy(SchemaObjectIdent('', row['database'], row['schema'], row['name']))

        return ResolveResult.DROP

    def _create_policy(self, bp: MaskingPolicyBlueprint):
        query = self.engine.query_builder()

        query.append("CREATE MASKING POLICY {full_name:i} AS (", {
            "full_name": bp.full_name,
        })

        for idx, arg in enumerate(bp.arguments):
            query.append_nl("    {comma:r}{arg_name:i} {arg_type:r}", {
                "comma": "  " if idx == 0 else ", ",
                "arg_name": arg.name,
                "arg_type": arg.type,
            })

        query.append_nl(")")

        query.append_nl("RETURNS {ret_type:r} -> ", {
            "ret_type": bp.returns,
        })

        query.append_nl("{body:r}", {
            "body": bp.body,
        })

        if bp.comment:
            query.append_nl("COMMENT = {comment}", {
                "comment": bp.comment,
            })

        self.engine.execute_unsafe_ddl(query, condition=self.engine.settings.execute_masking_policy)

    def _drop_policy(self, policy: SchemaObjectIdent):
        self.engine.execute_unsafe_ddl("DROP MASKING POLICY {full_name:i}", {
            "full_name": policy
        }, condition=self.engine.settings.execute_masking_policy)

    def _apply_policy_refs(self, bp: MaskingPolicyBlueprint, skip_existing=False):
        existing_policy_refs = {} if skip_existing else self._get_existing_policy_refs(bp.full_name)
        applied_change = False

        for ref in bp.references:
            ref_key = f"{ref.object_type.name}|{ref.object_name}|{ref.columns[0]}"

            # Policy was applied before
            if ref_key in existing_policy_refs:
                del(existing_policy_refs[ref_key])
                continue

            # Apply new masking policy
            self.engine.execute_unsafe_ddl("ALTER {object_type:r} {object_name:i} MODIFY COLUMN {first_column:i} SET MASKING POLICY {policy_name:i} USING ({columns:i})", {
                "object_type": ref.object_type.singular,
                "object_name": ref.object_name,
                "policy_name": bp.full_name,
                "first_column": ref.columns[0],
                "columns": ref.columns,
            }, condition=self.engine.settings.execute_masking_policy)

            applied_change = True

        # Remove remaining policy references which no longer exist in blueprint
        for existing_ref in existing_policy_refs.values():
            self.engine.execute_unsafe_ddl("ALTER {object_type:r} {database:i}.{schema:i}.{name:i} MODIFY COLUMN {first_column:i} UNSET MASKING POLICY", {
                "object_type": existing_ref['object_type'],
                "database": existing_ref['database'],
                "schema": existing_ref['schema'],
                "name": existing_ref['name'],
                "first_column": existing_ref['first_column'],
            }, condition=self.engine.settings.execute_masking_policy)

            applied_change = True

        return applied_change

    def _drop_policy_refs(self, policy_name: SchemaObjectIdent):
        existing_policy_refs = self._get_existing_policy_refs(policy_name)

        for existing_ref in existing_policy_refs.values():
            self.engine.execute_unsafe_ddl("ALTER {object_type:r} {database:i}.{schema:i}.{name:i} MODIFY COLUMN {first_column:i} UNSET MASKING POLICY", {
                "object_type": existing_ref['object_type'],
                "database": existing_ref['database'],
                "schema": existing_ref['schema'],
                "name": existing_ref['name'],
                "first_column": existing_ref['first_column'],
            }, condition=self.engine.settings.execute_masking_policy)

    def _get_existing_policy_refs(self, policy_name: SchemaObjectIdent):
        existing_policy_refs = {}

        cur = self.engine.execute_meta("SELECT * FROM TABLE({database:i}.information_schema.policy_references(policy_name => {policy_name}))", {
            "database": policy_name.database_full_name,
            "policy_name": policy_name,
        })

        for r in cur:
            ref_key = f"{r['REF_ENTITY_DOMAIN']}|{r['REF_DATABASE_NAME']}.{r['REF_SCHEMA_NAME']}.{r['REF_ENTITY_NAME']}|{r['REF_COLUMN_NAME']}"

            existing_policy_refs[ref_key] = {
                "object_type": r['REF_ENTITY_DOMAIN'],
                "database": r['REF_DATABASE_NAME'],
                "schema": r['REF_SCHEMA_NAME'],
                "name": r['REF_ENTITY_NAME'],
                "first_column": r['REF_COLUMN_NAME'],
            }

        return existing_policy_refs



from argparse import ArgumentParser, HelpFormatter
from json import loads as json_loads
from json.decoder import JSONDecodeError
from logging import getLogger, Formatter, StreamHandler
from os import environ, getcwd
from pathlib import Path
from snowflake.connector import connect
from traceback import TracebackException

from snowddl.blueprint import ObjectType
from snowddl.config import SnowDDLConfig
from snowddl.engine import SnowDDLEngine
from snowddl.parser import default_parser_sequence, PlaceholderParser
from snowddl.resolver import default_resolver_sequence
from snowddl.settings import SnowDDLSettings
from snowddl.version import __version__


class BaseApp:
    parser_sequence = default_parser_sequence
    resolver_sequence = default_resolver_sequence

    def __init__(self):
        self.arg_parser = self.init_arguments_parser()
        self.args = self.init_arguments()
        self.logger = self.init_logger()

        self.config_path = self.init_config_path()
        self.config = self.init_config()
        self.settings = self.init_settings()

        self.engine = self.init_engine()

    def init_arguments_parser(self):
        formatter = lambda prog: HelpFormatter(prog, max_help_position=32)
        parser = ArgumentParser(prog='snowddl', description='Object management automation tool for Snowflake', formatter_class=formatter)

        # Config
        parser.add_argument('-c', help='Path to config directory OR name of bundled test config (default: current directory)', metavar='CONFIG_PATH', default=getcwd())

        # Auth
        parser.add_argument('-a', help='Snowflake account identifier (default: SNOWFLAKE_ACCOUNT env variable)', metavar='ACCOUNT', default=environ.get('SNOWFLAKE_ACCOUNT'))
        parser.add_argument('-u', help='Snowflake user name (default: SNOWFLAKE_USER env variable)', metavar='USER', default=environ.get('SNOWFLAKE_USER'))
        parser.add_argument('-p', help='Snowflake user password (default: SNOWFLAKE_PASSWORD env variable)', metavar='PASSWORD', default=environ.get('SNOWFLAKE_PASSWORD'))
        parser.add_argument('-k', help='Path to private key file (default: SNOWFLAKE_PRIVATE_KEY_PATH env variable)', metavar='PRIVATE_KEY', default=environ.get('SNOWFLAKE_PRIVATE_KEY_PATH'))

        # Role & warehouse
        parser.add_argument('-r', help='Snowflake active role (default: SNOWFLAKE_ROLE env variable)', metavar='ROLE', default=environ.get('SNOWFlAKE_ROLE'))
        parser.add_argument('-w', help='Snowflake active warehouse (default: SNOWFLAKE_WAREHOUSE env variable)', metavar='WAREHOUSE', default=environ.get('SNOWFLAKE_WAREHOUSE'))

        # Options
        parser.add_argument('--passphrase', help='Passphrase for private key file (default: SNOWFLAKE_PRIVATE_KEY_PASSPHRASE env variable)', default=environ.get('SNOWFLAKE_PRIVATE_KEY_PASSPHRASE'))
        parser.add_argument('--env-prefix', help='Env prefix added to global object names, used to separate environments (e.g. DEV, PROD)', default=None)
        parser.add_argument('--max-workers', help='Maximum number of workers to resolve objects in parallel', default=None, type=int)

        # Logging
        parser.add_argument('--log-level', help="Log level (possible values: DEBUG, INFO, WARNING; default: INFO)", default="INFO")
        parser.add_argument('--show-sql', help="Show executed DDL queries", default=False, action='store_true')

        # Placeholders
        parser.add_argument('--placeholder-path', help='Path to config file with environment-specific placeholders', default=None, metavar='')
        parser.add_argument('--placeholder-values', help='Environment-specific placeholder values in JSON format', default=None, metavar='')

        # Object types
        parser.add_argument('--exclude-object-types', help="Comma-separated list of object types NOT to resolve", default=None, metavar='')
        parser.add_argument('--include-object-types', help="Comma-separated list of object types TO resolve, all other types are excluded", default=None, metavar='')

        # Apply even more unsafe changes
        parser.add_argument('--apply-unsafe', help="Additionally apply unsafe changes, which may cause loss of data (ALTER, DROP, etc.)", default=False, action='store_true')
        parser.add_argument('--apply-replace-table', help="Additionally apply REPLACE TABLE when ALTER TABLE is not possible", default=False, action='store_true')
        parser.add_argument('--apply-masking-policy', help="Additionally apply changes to MASKING POLICIES", default=False, action='store_true')
        parser.add_argument('--apply-row-access-policy', help="Additionally apply changes to ROW ACCESS POLICIES", default=False, action='store_true')
        parser.add_argument('--apply-account-params', help="Additionally apply changes to ACCOUNT PARAMETERS", default=False, action='store_true')
        parser.add_argument('--apply-network-policy', help="Additionally apply changes to NETWORK POLICIES", default=False, action='store_true')
        parser.add_argument('--apply-resource-monitor', help="Additionally apply changes to RESOURCE MONITORS", default=False, action='store_true')
        # parser.add_argument('--apply-inbound-share', help="Additionally apply changes to INBOUND SHARES", default=False, action='store_true')
        parser.add_argument('--apply-outbound-share', help="Additionally apply changes to OUTBOUND SHARES", default=False, action='store_true')

        # Refresh state of specific objects
        parser.add_argument('--refresh-user-passwords', help="Additionally refresh passwords of users", default=False, action='store_true')
        parser.add_argument('--refresh-future-grants', help="Additionally refresh missing grants for existing objects derived from future grants", default=False, action='store_true')

        # Destroy without env prefix
        parser.add_argument('--destroy-without-prefix', help="Allow {destroy} action without --env-prefix", default=False, action='store_true')

        # Subparsers
        subparsers = parser.add_subparsers(dest="action")
        subparsers.required = True

        subparsers.add_parser('plan', help="Resolve objects, apply nothing, display suggested changes")
        subparsers.add_parser('apply', help="Resolve objects, apply safe changes, display suggested unsafe changes")
        subparsers.add_parser('destroy', help="Drop objects with specified --env-prefix, use it to reset dev and test environments")

        return parser

    def init_arguments(self):
        args = vars(self.arg_parser.parse_args())

        if not args['a'] or not args['u'] or (not args['p'] and not args['k']):
            self.arg_parser.print_help()
            exit(1)

        return args

    def init_logger(self):
        logger = getLogger('snowddl')
        logger.setLevel(self.args.get('log_level', 'INFO'))

        formatter = Formatter('%(asctime)s - %(levelname)s - %(message)s')
        formatter.default_msec_format = '%s.%03d'

        handler = StreamHandler()
        handler.setFormatter(formatter)

        logger.addHandler(handler)

        return logger

    def init_config_path(self):
        config_path = Path(self.args['c'])

        if not config_path.exists():
            config_path = Path(__file__).parent.parent / '_config' / self.args['c']

        if not config_path.exists():
            raise ValueError(f"Config path [{self.args['c']}] does not exist")

        if not config_path.is_dir():
            raise ValueError(f"Config path [{self.args['c']}] is not a directory")

        return config_path.resolve()

    def init_config(self):
        config = SnowDDLConfig(self.args.get('env_prefix'))

        # Placeholders
        placeholder_path = self.get_placeholder_path()
        placeholder_values = self.get_placeholder_values()

        parser = PlaceholderParser(config, self.config_path)
        parser.load_placeholders(placeholder_path, placeholder_values)

        if config.errors:
            self.output_config_errors(config)
            exit(1)

        # Blueprints
        for parser_cls in self.parser_sequence:
            parser = parser_cls(config, self.config_path)
            parser.load_blueprints()

        if config.errors:
            self.output_config_errors(config)
            exit(1)

        return config

    def init_settings(self):
        settings = SnowDDLSettings()

        if self.args.get('action') in ('apply', 'destroy'):
            settings.execute_safe_ddl = True

            if self.args.get('apply_unsafe') or self.args.get('action') == 'destroy':
                settings.execute_unsafe_ddl = True

            if self.args.get('apply_replace_table'):
                settings.execute_replace_table = True

            if self.args.get('apply_masking_policy'):
                settings.execute_masking_policy = True

            if self.args.get('apply_row_access_policy'):
                settings.execute_row_access_policy = True

            if self.args.get('apply_account_params'):
                settings.execute_account_params = True

            if self.args.get('apply_network_policy'):
                settings.execute_network_policy = True

            if self.args.get('apply_resource_monitor'):
                settings.execute_resource_monitor = True

            if self.args.get('apply_inbound_share'):
                settings.execute_inbound_share = True

            if self.args.get('apply_outbound_share'):
                settings.execute_outbound_share = True

        if self.args.get('refresh_user_passwords'):
            settings.refresh_user_passwords = True

        if self.args.get('refresh_future_grants'):
            settings.refresh_future_grants = True

        if self.args.get('exclude_object_types'):
            try:
                settings.exclude_object_types = [ObjectType[t.strip().upper()] for t in str(self.args.get('exclude_object_types')).split(',')]
            except KeyError as e:
                raise ValueError(f"Invalid object type [{str(e)}]")

        if self.args.get('include_object_types'):
            try:
                settings.include_object_types = [ObjectType[t.strip().upper()] for t in str(self.args.get('include_object_types')).split(',')]
            except KeyError as e:
                raise ValueError(f"Invalid object type [{str(e)}]")

        if self.args.get('max_workers'):
            settings.max_workers = int(self.args.get('max_workers'))

        return settings

    def init_engine(self):
        return SnowDDLEngine(self.get_connection(), self.config, self.settings)

    def get_connection(self):
        options = {
            "account": self.args['a'],
            "user": self.args['u'],
            "role": self.args['r'],
            "warehouse": self.args['w'],
        }

        if self.args.get('k'):
            from cryptography.hazmat.primitives import serialization

            key_path = Path(self.args.get('k'))
            key_password = str(self.args.get('passphrase')).encode('utf-8') if self.args.get('passphrase') else None

            pk = serialization.load_pem_private_key(data=key_path.read_bytes(), password=key_password)

            options['private_key'] =  pk.private_bytes(
                encoding=serialization.Encoding.DER,
                format=serialization.PrivateFormat.PKCS8,
                encryption_algorithm=serialization.NoEncryption(),
            )
        else:
            options['password'] = self.args['p']

        return connect(**options)

    def execute(self):
        error_count = 0

        with self.engine:
            self.output_engine_context()

            if self.args.get('action') == 'destroy':
                if not self.args.get('env_prefix') and not self.args.get('destroy_without_prefix'):
                    raise ValueError("Argument --env-prefix is required for [destroy] action")

                for resolver_cls in self.resolver_sequence:
                    resolver = resolver_cls(self.engine)
                    resolver.destroy()

                    error_count += len(resolver.errors)

                self.engine.context.destroy_role_with_prefix()
            else:
                for resolver_cls in self.resolver_sequence:
                    resolver = resolver_cls(self.engine)
                    resolver.resolve()

                    error_count += len(resolver.errors)

            self.output_engine_stats()

            if self.args.get('show_sql'):
                self.output_executed_ddl()

            self.output_suggested_ddl()

            if error_count > 0:
                exit(8)

    def output_engine_context(self):
        roles = []

        if self.engine.context.is_account_admin:
            roles.append("ACCOUNTADMIN")

        if self.engine.context.is_sys_admin:
            roles.append("SYSADMIN")

        if self.engine.context.is_security_admin:
            roles.append("SECURITYADMIN")

        self.logger.info(f"Snowflake version = {self.engine.context.version} ({self.engine.context.edition.name}), SnowDDL version = {__version__}")
        self.logger.info(f"Session = {self.engine.context.current_session}, User = {self.engine.context.current_user}")
        self.logger.info(f"Role = {self.engine.context.current_role}, Warehouse = {self.engine.context.current_warehouse}")
        self.logger.info(f"Roles in session = {','.join(roles)}")
        self.logger.info("---")

    def get_placeholder_path(self):
        if self.args.get('placeholder_path'):
            placeholder_path = Path(self.args.get('placeholder_path'))

            if not placeholder_path.is_file():
                raise ValueError(f"Placeholder path [{self.args.get('placeholder_path')}] does not exist or not a file")

            return placeholder_path.resolve()

        return None

    def get_placeholder_values(self):
        if self.args.get('placeholder_values'):
            try:
                placeholder_values = json_loads(self.args.get('placeholder_values'))
            except JSONDecodeError as e:
                raise ValueError(f"Placeholder values [{self.args.get('placeholder_values')}] are not a valid JSON")

            if not isinstance(placeholder_values, dict):
                raise ValueError(f"Placeholder values [{self.args.get('placeholder_values')}] are not JSON encoded dict")

            for k, v in placeholder_values.items():
                if not isinstance(v, (bool, float, int, str)):
                    raise ValueError(f"Invalid type [{type(v)}] of placeholder [{k.upper()}] value, supported types are: bool, float, int, str")

            return placeholder_values

        return None

    def output_config_errors(self, config):
        for e in config.errors:
            self.logger.warning(f"[{e['path']}]: {''.join(TracebackException.from_exception(e['error']).format())}")

    def output_engine_stats(self):
        self.logger.info(f"Executed {len(self.engine.executed_ddl)} DDL queries, Suggested {len(self.engine.suggested_ddl)} DDL queries")

    def output_suggested_ddl(self):
        if self.engine.suggested_ddl:
            print("--- Suggested DDL ---\n")

        for sql in self.engine.suggested_ddl:
            print(f"{sql};\n")

    def output_executed_ddl(self):
        if self.engine.executed_ddl:
            print("--- Executed DDL ---\n")

        for sql in self.engine.executed_ddl:
            print(f"{sql};\n")


def entry_point():
    app = BaseApp()
    app.execute()
